/**
 * Copyright (c) 2016-2018 Thales Global Services S.A.S.
 * This program and the accompanying materials are made
 * available under the terms of the Eclipse Public License 2.0
 * which is available at https://www.eclipse.org/legal/epl-2.0/
 * 
 * SPDX-License-Identifier: EPL-2.0
 * 
 * Contributors:
 *   Abel Hegedus, Tamas Borbas, Peter Lunk, Daniel Segesdi (IncQuery Labs Ltd.) - initial API and implementation
 */
package org.eclipse.emf.diffmerge.patch.api;

import com.google.common.base.Objects;
import java.util.Arrays;
import java.util.function.Consumer;
import org.eclipse.emf.diffmerge.patch.AttributeEntry;
import org.eclipse.emf.diffmerge.patch.ElementEntry;
import org.eclipse.emf.diffmerge.patch.ReferenceEntry;
import org.eclipse.emf.diffmerge.patch.api.ChangeDirection;
import org.eclipse.emf.diffmerge.patch.api.Identifiable;
import org.eclipse.emf.diffmerge.patch.api.ModelPatch;
import org.eclipse.emf.diffmerge.patch.api.ModelPatchBuilder;
import org.eclipse.emf.diffmerge.patch.api.ModelPatchEntry;
import org.eclipse.emf.diffmerge.patch.api.ModelPatchEntryBuilder;
import org.eclipse.emf.diffmerge.patch.api.ModelPatchException;
import org.eclipse.xtext.xbase.lib.Exceptions;
import org.eclipse.xtext.xbase.lib.ListExtensions;
import org.eclipse.xtext.xbase.lib.ObjectExtensions;
import org.eclipse.xtext.xbase.lib.Procedures.Procedure1;

@SuppressWarnings("all")
public class ModelPatchReverser {
  public static final ModelPatchReverser INSTANCE = new ModelPatchReverser();
  
  public ModelPatch reverse(final ModelPatch originalPatch) throws ModelPatchException {
    final ModelPatchBuilder reverseBuilder = ModelPatchBuilder.create();
    final Consumer<ModelPatchEntry> _function = (ModelPatchEntry originalEntry) -> {
      try {
        reverseBuilder.addEntry(this.reverseEntry(originalEntry));
      } catch (Throwable _e) {
        throw Exceptions.sneakyThrow(_e);
      }
    };
    ListExtensions.<ModelPatchEntry>reverseView(originalPatch.getEntries()).forEach(_function);
    return reverseBuilder.build();
  }
  
  public ModelPatchEntry _reverseEntry(final ModelPatchEntry originalEntry) throws ModelPatchException {
    throw new ModelPatchException("Unhandled entry type");
  }
  
  public ModelPatchEntry _reverseEntry(final AttributeEntry originalEntry) {
    ModelPatchEntryBuilder _reverseEntryBuilder = this.reverseEntryBuilder(originalEntry);
    final Procedure1<ModelPatchEntryBuilder> _function = (ModelPatchEntryBuilder it) -> {
      it.setFeature(this.cloneIdentifiable(originalEntry.getFeature()));
      it.setValue(originalEntry.getValue());
      it.setIndex(originalEntry.getIndex());
    };
    final ModelPatchEntryBuilder reverseEntryBuilder = ObjectExtensions.<ModelPatchEntryBuilder>operator_doubleArrow(_reverseEntryBuilder, _function);
    return reverseEntryBuilder.buildAttributeEntry();
  }
  
  public ModelPatchEntry _reverseEntry(final ElementEntry originalEntry) {
    ModelPatchEntryBuilder _reverseEntryBuilder = this.reverseEntryBuilder(originalEntry);
    final Procedure1<ModelPatchEntryBuilder> _function = (ModelPatchEntryBuilder it) -> {
      it.setType(this.cloneIdentifiable(originalEntry.getType()));
    };
    final ModelPatchEntryBuilder reverseEntryBuilder = ObjectExtensions.<ModelPatchEntryBuilder>operator_doubleArrow(_reverseEntryBuilder, _function);
    return reverseEntryBuilder.buildElementEntry();
  }
  
  public ModelPatchEntry _reverseEntry(final ReferenceEntry originalEntry) {
    ModelPatchEntryBuilder _reverseEntryBuilder = this.reverseEntryBuilder(originalEntry);
    final Procedure1<ModelPatchEntryBuilder> _function = (ModelPatchEntryBuilder it) -> {
      it.setFeature(this.cloneIdentifiable(originalEntry.getFeature()));
      it.setTarget(this.cloneIdentifiable(originalEntry.getTarget()));
      it.setIndex(originalEntry.getIndex());
    };
    final ModelPatchEntryBuilder reverseEntryBuilder = ObjectExtensions.<ModelPatchEntryBuilder>operator_doubleArrow(_reverseEntryBuilder, _function);
    return reverseEntryBuilder.buildReferenceEntry();
  }
  
  public ChangeDirection reverseDirection(final ChangeDirection original) {
    boolean _equals = Objects.equal(ChangeDirection.ADD, original);
    if (_equals) {
      return ChangeDirection.REMOVE;
    } else {
      return ChangeDirection.ADD;
    }
  }
  
  private ModelPatchEntryBuilder reverseEntryBuilder(final ModelPatchEntry originalEntry) {
    return ModelPatchBuilder.entryBuilder(
      this.cloneIdentifiable(originalEntry.getContext()), 
      this.reverseDirection(originalEntry.getDirection()));
  }
  
  private Identifiable cloneIdentifiable(final Identifiable original) {
    String _identifier = original.getIdentifier();
    return new Identifiable(_identifier);
  }
  
  public ModelPatchEntry reverseEntry(final ModelPatchEntry originalEntry) throws ModelPatchException {
    if (originalEntry instanceof AttributeEntry) {
      return _reverseEntry((AttributeEntry)originalEntry);
    } else if (originalEntry instanceof ReferenceEntry) {
      return _reverseEntry((ReferenceEntry)originalEntry);
    } else if (originalEntry instanceof ElementEntry) {
      return _reverseEntry((ElementEntry)originalEntry);
    } else if (originalEntry != null) {
      return _reverseEntry(originalEntry);
    } else {
      throw new IllegalArgumentException("Unhandled parameter types: " +
        Arrays.<Object>asList(originalEntry).toString());
    }
  }
}
